;Copyright 2007 Bueno Systems, Inc.
;
;Permission is hereby granted, free of charge, to any person obtaining a copy of this software
;and associated documentation files (the "Software"), to deal in the Software without restriction,
;including without limitation the rights to use, copy, modify, merge, publish, distribute,
;sublicense, and/or sell copies of the Software,and to permit persons to whom the Software is
;furnished to do so, subject to the following conditions:
;
;The above copyright notice and this permission notice shall be included in all copies or
;substantial portions of the Software.
;
;THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
;INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
;AND NONINFRINGEMENT. IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
;CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
;
	.NOLIST
	.DEVICE	ATTiny13
	.INCLUDE	"TN13DEF.INC"
	.INCLUDE "\jobs\avr\BSIMACROS.INC"
	.LIST
	.LISTMAC

;--------Equates.----------------------------

	.EQU	OWIO	= PINB2
	.EQU	SI	= PINB4
	.EQU	CLK	= PINB1
	.EQU	LEDPWM	= PINB0

	.EQU	TRIG	= CLK

	.EQU	DEVICE0	= 'L'
	.EQU	DEVICE1	= 'S'
	.EQU	VERSION	= 1

;--------Registers.--------------------------

	.DEF	MINLOC	= R10
	.DEF	MAXLOC	= R11
	.DEF	SHIFT	= R12
	.DEF	EXPTIME	= R13
	.DEF	THLD	= R14
	.DEF	HYST	= R15

	.DEF	LED	= R18
	.DEF	MACREG	= R19
	.DEF	ACC	= R20
	.DEF	FILTER	= R21
	.DEF	COMMAND	= R22
	.DEF	PTIME	= R23
	.DEF	BITCNT	= R24
	.DEF	FLAGS	= R25

	.EQU	STROBE	= 7		;In LED.

	.EQU	BADCMD	= 7		;In FLAGS.
	.EQU	CANTBUF	= 6
	.EQU	CMDOVF	= 5
	.EQU	DATOVF	= 4
	.EQU	BUFING	= 3
	.EQU	XEQING	= 2

	.EQU	FLOAT	= 7		;In FILTER.
	.EQU     WINDOW	= 6

	.EQU	XTRIG	= 3		;In COMMAND for Acquire.
	.EQU	NOTPIX	= 2
	.EQU	ORPIX	= 1
	.EQU	ANDPIX	= 0

	.EQU	NEW	= 3		;In COMMAND for Count and Find.
	.EQU	REV	= 2
	.EQU	BRT	= 1
	.EQU	EDG	= 0

;--------SRAM.-------------------------------

	.DSEG

	.ORG	0x60

PIXELS:	.BYTE	32
RESULTS:	.BYTE	5
CMDBUF:	.BYTE	11
MAXRAM:

;--------Code.-------------------------------

	.CSEG

	.ORG	0

;--------Start here.-------------------------	

START:	CLR	R0
	OUT	OCR0A, R0 		;Initialize LED strobe.
	OUTI	TCCR0B,0b00000001		;Set PWM clock to max.
	OUTI	TCCR0A,0b00000011		;Fast PWM on OC0A, but disabled.
	OUT	PORTB, R0		;Clear port B.
	OUTI	DDRB,  0b00010011		;Enable outputs on SI, Clk, and LED.
	OUTI	DIDR0, 0b00001000		;Disable digital input on analog pin.
	OUTI	ADCSRA,0b10000100		;Enable the ADC with /64 clock.
	OUTI	ADMUX, 0b00100011		;Select PB.3 with Vdd ref, left-adjusted.
	OUTI	PCMSK, 0b00000100		;Interrupt on OWIO only.
	OUTI	GIMSK, 0b00100000		;Enable interrupt on pin change.
	LDI	ZL,0x60			;Point Z to beginning of RAM.

CLRLP:	ST	Z,R0			;Clear RAM & Reg locations, but not I/O space.
	INC	ZL			;Point to next. (No Z+ for no carry after 0xFF.)
	CPI	ZL,0x1E			;ZL points to ZL?
	BRNE	CLRLP			;  No:  Branch back.

	LDI	MACREG,128
	MOV	THLD,MACREG
	MOV	EXPTIME,MACREG

;--------Protocol reset----------------------

RESET:	OUTI	SPL,RAMEND		;Stamp OW-reset jumps here: Clear stack.
	ANDI	FLAGS,0xFF-(1<<XEQING)-(1<<BUFING)

;--------Change buffer/execution state-------

CHGST:	LDI	XL,RESULTS		;State change: re-init pointers.
	LDI	YL,CMDBUF

;--------Main program loop-------------------


MAINLP:	RCALL	GETNEXT			;Get the next command.

GOTCMD:	MOV	COMMAND,ACC		;Save in COMMAND.
	JBRC	FLAGS,BUFING,DOIT		;Are we just buffering for now?

	RCALL	BUFCMD			;  Yes: Buffer the command.
	CPIBE	ACC,0x3E,GO		;       Time for action if '>'.
	ANDI	ACC,0xF7
	CPIBE	ACC,0xA0,NOBUF
	ANDI	ACC,0xF0			;       Get command MSN.
	CPIBE	ACC,0xA0,MAINLP		;       Acquire can be buffered ... maybe.
	CPIBE	ACC,0xC0,COUNT		;       Count can be buffered.
	CPIBE	ACC,0xD0,NOBUF		;       Dump cannot be buffered.
	CPIBE	ACC,0xE0,ENTER		;       Enter can be buffered.
	CPIBE	ACC,0xF0,COUNT		;       Find can be buffered. Has same form as Count.
	RJMP	NOTACMD

NOBUF:	ORI	FLAGS,1<<CANTBUF		;Can't buffer. Flag the error.
	RJMP	ERROR

COUNT:	JBRC	COMMAND,NEW,MAINLP	;Count/Find reading new args?
	RJMP	BUF2			;  Yes: Two of them. Get and buffer.

ENTER:	MOV	ACC,COMMAND		;Get command.
	CPIBE	ACC,0xEB,BUF1		;LED brightness setting uses one arg.
	CPIBE	ACC,0xEC,BUF3		;Binary coefficient settings use three args.
	CPIBE	ACC,0xEE,BUF1		;Exposure time is one arg.
	RJMP	NOTACMD			;Invalid command, otherwise.

BUF3:	RCALL	GETNBUF			;Buffer 3 bytes.
BUF2:	RCALL	GETNBUF			;Buffer 2 bytes.
BUF1:	RCALL	GETNBUF			;Buffer 1 byte.
	RJMP	MAINLP			;Back for another command.

GO:	ORI	FLAGS,1<<XEQING		;Set "execute from buffer" flag.
	ANDI	FLAGS,0xFF - (1<<BUFING)	;Clear "buffering commands" flag.
	RJMP	CHGST			;Go change state.

DOIT:	CPIBE	ACC,'<',BUFFER		;Executing direct.
	CPIBE	ACC,'>',STOP		;STOP if end of buffer.
	ANDI	ACC,0xF0			;Get command MSN.
	CPIJE	ACC,0xA0,DOACQ		;A = Acquire pixels.
	CPIBE	ACC,0xC0,DOCOUNT		;C = Count pixels.
	CPIBE	ACC,0xD0,DODUMP		;D = Dump data.
	CPIJE	ACC,0xE0,DOENTER		;E = Enter data.
	CPIJE	ACC,0xF0,DOFIND		;F = Find pixels.
	RJMP	NOTACMD			;Invalid command, otherwise.

BUFFER:	ORI	FLAGS,1<<BUFING		;Starting to buffer. Set "buffering commands" flag.
	RJMP	CHGST			;Go change state.

STOP:	RCALL	PUTRDY			;End of buffer. Signal we're done.
	ANDI	FLAGS,0xFF - (1<<XEQING)	;Clear "executing from buffer" flag.
	RJMP	CHGST			;Go change state.

DOCOUNT:					;Count pixels or edges.
	
	.DEF	LSAVE   = R0		;Save MINLOC here for later restore.
	.DEF	RSAVE   = R1		;Save MAXLOC here for later restore.
	.DEF	COUNTER = R2		;Feature counter.

	CLR	COUNTER			;Clear pixel counter.
	ANDI	COMMAND,0xFF - (1<<REV)	;Force count to go forward.
	RCALL	FNDPIX			;Find first pixel/edge.
	JBRS	COMMAND,EDG,CNTEDG	;Counting edges?

CNTPIX:	BRTC	CNTDONE			;  No:  Pixels. T cleared means no more in range.

	SUB	COUNTER,MINLOC		;       Subtract first pixel position of desired polarity.
	RCALL	FNDNXT			;       Find next pixel of opposite polarity.
	ADD	COUNTER,MINLOC		;       Add its position to counter. (Adds MAXLOC + 1 if not found.)
	BRTC	CNTDONE			;       Done if opp not found.

	RCALL	FNDNXT			;       Find next pixel of desired polarity.
	RJMP	CNTPIX			;       Do it some more...

CNTEDG:	BRTC	CNTDONE			;  Yes: Edges. T cleared means no more in range.

	INC	COUNTER			;       T set means found an edge. Increment count.
	RCALL	FNDNXT			;       Find the next edge.
	RJMP	CNTEDG			;       Do it some more...

CNTDONE:	MOV	ACC,COUNTER		;Done counting. Buffer the result.
	RCALL	BUFDAT
	MOVW	MAXLOC:MINLOC,RSAVE:LSAVE	;Restore pointers.
	RJMP	CNDRDY			;Send ready if direct execution; loop back to MAINLP.

DODUMP:	MOV	ACC,COMMAND		;Dump data to host. Get command.	
	CPIBE	ACC,0xDA,DUMPADR		;Dump from address.
	CPIBE	ACC,0xDD,DUMPID		;Dump firmware ID.
	CPIBE	ACC,0xDF,DUMPFLG		;Dump flags.
	RJMP	NOTACMD
	
DUMPADR:	RCALL	GETBYTE			;Get address of first byte to dump.
	SUBI	ACC,-PIXELS		;Adding PIXELS results in overflow?
	BRCS	DMPOK

	RJMP	ERROR			;  Yes: Wait for reset.

DMPOK:	MOV	ZL,ACC			;  No:  Init pointer.

DUMPLP:	CPI	ZL,MAXRAM		;Pointer beyond dumpable area?
	BRCS	DMPLPOK

	RJMP	ERROR			;  Yes: Wait for reset.

DMPLPOK:	LD	ACC,Z+			;  No:  Get next datum.
	RCALL	PUTBYTE			;       Send it.
	RJMP	DUMPLP			;       Back for another.

DUMPID:	LDI	ACC,DEVICE0		;Dump firmware version info.
	RCALL	PUTBYTE
	LDI	ACC,DEVICE1
	RCALL	PUTBYTE
	LDI	ACC,VERSION
	RCALL	PUTBYTE
	RJMP	MAINLP			;Back for another command.

DUMPFLG:	MOV	ACC,FLAGS		;Dump flags.
	ANDI	ACC,0xF0			;Just the error flags.
	RCALL	PUTBYTE
	ANDI	FLAGS,0xFF-(1<<BADCMD)-(1<<CANTBUF)-(1<<CMDOVF)-(1<<DATOVF)	;Then clear errors.
	RJMP	MAINLP			;Back for another command.

DOENTER:	MOV	ACC,COMMAND		;Enter settings. Get command.
	CPIBE	ACC,0xEB,ENTLED		;LED brightness setting.
	CPIBE	ACC,0xEC,ENTBIN		;Binary coefficient settings.
	CPIBE	ACC,0xEE,ENTEXP		;Exposure time.
	RJMP	NOTACMD			;Otherwise, invalid command.

ENTBIN:	RCALL	GETNEXT			;Enter binary filter settings:
	MOV	THLD,ACC			;Threshold,
	RCALL	GETNEXT
	MOV	HYST,ACC			;  hysteresis,
	RCALL	GETNEXT
	MOV	FILTER,ACC		;  and filter.
	RJMP	MAINLP			;Back for another command.

ENTEXP:	RCALL	GETNEXT			;Enter exposure time.
	TST	ACC			;Is it > 0?
	BRNE	SETEXP			

	INC	ACC			;  No:  Set to 1.

SETEXP:	MOV	EXPTIME,ACC		;Save the value.
	RJMP	MAINLP			;Back for another command.

ENTLED:	RCALL	GETNEXT			;Enter LED setting.
	MOV	LED,ACC			;Save in register.
	SBRC	LED,STROBE		;Strobing?
	SER	ACC			;  Yes: Set LED for maximum.
	OUT	OCR0A,ACC		;Write brightness value to PWM.
	RJMP	MAINLP			;Back for another command.

DOFIND:	RCALL	FNDPIX			;Find pixels or edges in binary scan.
	INC	ACC			;Returned value is 1-based. Internally it's 0-based.
	RCALL	BUFDAT			;Buffer the result.
	RJMP	CNDRDY			;Send ready if executing directly; loop back to MAINLP.

FNDPIX:	JBRC	COMMAND,NEW,ARGSOK	;Does command require new search limits?

	RCALL	GETNEXT			;  Yes: Get lower pixel position.
	MOV	MINLOC,ACC
	DEC	MINLOC			;       User value is 1-based; internally 0-based.
	RCALL	GETNEXT
	MOV	MAXLOC,ACC		;       Get upper pixel position.
	DEC	MAXLOC			;       Same-o same-o.

ARGSOK:	MOVW	RSAVE:LSAVE,MAXLOC:MINLOC	;Save pointers, in case call was from CNTPIX.
	RCALL	FSTPIX			;Find first desired pixel.
	RJMP	TSTEDG			;Branch around.

FNDNXT:	RCALL	NXTPIX			;Find next desired pixel.

TSTEDG:	BRTC	FNDEND
	SBRC	COMMAND,EDG		;Looking for an edge?
	RCALL	NXTPIX			;  Yes: Find next pixel of opposite polarity.

FNDEND:	MOV	ACC,MINLOC		;  First Pixel: Return MINLOC.
	SBRC	COMMAND,REV
	MOV	ACC,MAXLOC		;  Last Pixel:  Return MAXLOC.
	BRTS	FNDOUT

	SER	ACC
FNDOUT:	RET				;Over and out.

FSTPIX:	MOV	ZL,MINLOC		;Start at MINLOC conditionally.
	SBRC	COMMAND,REV		;Scanning in reverse?
	MOV	ZL,MAXLOC		;   Yes: Use MAXLOC instead.
	LSR	ZL			;Divide pixel no. by 8 to get buffer address.
	LSR	ZL
	LSR	ZL
	SUBI	ZL,-PIXELS		;Calculate RAM address
	SET				;Set "beginning pixel" flag.
	RJMP	LDPIX			;Go get 'em.

NXTPIX:	CLT				;Clear "beginning pixel" flag.
	JBRS	COMMAND,REV,NXTREV	;Get next pixel. Which way are we scanning?

NXTFWD:	INC	MINLOC			;  Fwd: Increment low pointer
	MOV	ACC,MINLOC		;       Get it.
	ANDI	ACC,0x07			;       Multiple of eight?
	BRNE	THSPIX			;         No:  Still pixels in SHIFT.
	INC	ZL			;         Yes: Point to next 8 pixels in buffer.
	RJMP	LDPIX			;              Go get 'em.

NXTREV:	MOV	ACC,MAXLOC		;  Rev: Get current high pointer.
	DEC	MAXLOC			;       Then decrement high pointer.
	ANDI	ACC,0x07			;       Was value multiple of eight?
	BRNE	THSPIX			;         No: Still pixels in SHIFT.
	DEC	ZL			;             Point to previous 8 pixels in buffer.

LDPIX:	LD	SHIFT,Z			;Get 8 more pixels.
	SBRS	COMMAND,BRT		;Looking for bright pixels?
	COM	SHIFT			;  No:  Dark ones. Invert bits.
	BRTC	THSPIX			;Beginning pixel?
	
	CLT				;  Yes: Clear flag.
	JBRS	COMMAND,REV,FSTREV	;       Searching which direction?

FSTFWD:	SBRC	MINLOC,2			;         Fwd: Shift left by MINLOC mod 8.
	SWAP	SHIFT
	SBRC	MINLOC,1
	LSr	SHIFT
	SBRC	MINLOC,1
	LSr	SHIFT
	SBRC	MINLOC,0
	LSr	SHIFT
	RJMP	THSPIX

FSTREV:	SBRS	MAXLOC,2			;         Rev: Shift right by 7 - (MAXLOC mod 8).
	SWAP	SHIFT
	SBRS	MAXLOC,1
	LSl	SHIFT
	SBRS	MAXLOC,1
	LSl	SHIFT
	SBRS	MAXLOC,0
	LSl	SHIFT

THSPIX:	LDI	ACC,255			;Pixel is in position in SHIFT.
	CP	MAXLOC,ACC		;MAXLOC = 255 (-1)?
	BREQ	NOTFND			;  Yes: Null result.
	CP	MAXLOC,MINLOC		;MINLOC > MAXLOC (i.e. OR MINLOC = 255)?
	BRCS	NOTFND			;  No:  Pointers are okay.

	SBRS	COMMAND,REV		;Scanning which direction?
	LSr	SHIFT			;  Fwd: Shift left.
	SBRC	COMMAND,REV
	LSl	SHIFT			;  Rev: Shift right.
	BRCC	NXTPIX			;Carry clear means pixel not found. Search some more.

	SET				;Found one: set "found it" flag.
	COM	SHIFT			;Shift to opposite polarity.
	LDI	ACC,1<<BRT		;Switch polarity in command, too.
	EOR	COMMAND,ACC

NOTFND:	RET				;Return to caller.

;--------Acquire new pixels from sensor.----

DOACQ:	.DEF	PIXEL	= R0
	.DEF	pPIXEL	= R1
	.DEF	MINPIX	= R2
	.DEF	MAXPIX	= R3
	.DEF	_THLD	= R4
	.DEF	_STBTMR	= R5
	.DEF	PIXSUML	= R6
	.DEF	PIXSUMH	= R7

	.DEF	PIXCNT	= R16

	CLR	PIXSUML
	CLR	PIXSUMH
	MOV	PTIME,EXPTIME		;Initialize exposure timer.

	JBRC	COMMAND,XTRIG,TRIGD

	CBI	DDRB,TRIG
WAITX1:	JBIC	PINB,TRIG,WAITX1
WAITX0:	JBIS	PINB,TRIG,WAITX0
	SBI	DDRB,TRIG

TRIGD:	MOV	ACC,LED
	ANDI	ACC,0x7F
	BREQ	LEDOK

	MOV	_STBTMR,ACC
	OUTI	TCCR0A,0b10000011
		
LEDOK:	SBI	PORTB,SI			;Raise the SI pin.
	OUTI	TCNT0,0			;Clear TMR0.
	OUTI	TIFR0,1<<TOV0		;Clear TMR0 overflow flag.

CLKLP:	LDI	PIXCNT,10		;Initialize time granularity = 10 * 256 / 9.6E6 = 266.6uS.
TMR0WT:	SBI	PORTB,CLK		;Raise CLK pin.
	CBI	PORTB,SI			;Lower SI pin if it's up.
	CBI	PORTB,CLK		;Lower CLK pin.
	IN	ACC,TIFR0		;Timer overflow?
	ANDI	ACC,1<<TOV0
	BREQ	TMR0WT			;  No: Go back and wait for it.

	OUT	TIFR0,ACC		;  Yes: Clear it
	LDI	ACC,0b00000011		;Tentatively load LED shutoff value.
	DEC	_STBTMR			;Still in strobe time?
	BRNE	CHKCNT			;  Yes: Branch around.
	
	SBRC	LED,STROBE		;  No:  Are we strobing?
	OUT	TCCR0A,ACC		;         Yes: Turn off LED.

CHKCNT:	DEC	PIXCNT			;       Decrement granularity counter.
	BRNE	TMR0WT			;       Back if not zero.

	DEC	PTIME			;       Decrement exposure timer.
	BRNE	CLKLP			;       Back if not zero.
	
	OUT	TCCR0A,ACC		;Done with exposure: Turn off LED.
	SBI	PORTB,SI			;Raise SI pin.
	SBI	PORTB,CLK		;Raise CLK.
	CLR	PIXCNT			;Clear pixel counter.
	CBI	PORTB,SI			;Lower SI.
	CBI	PORTB,CLK		;Lower CLK.
	SBI	ADCSRA,ADSC		;Start an A/D conversion.
	LDI	XL,0xE0			;Initialize XL (5 lsbs) to zero, with 7 in bit counter.
	;CLR	XH			;Clear XH.
	CLR	MAXPIX			;MAXPIX = 0.
	CLR	MINPIX			;MINPIX = 255.
	DEC	MINPIX

READLP:	JBIS	ADCSRA,ADSC,READLP	;Wait for ADC result.

	IN	PIXEL,ADCH		;Get the ADC value.
	SBI	PORTB,CLK		;Raise CLK for next pixel.
	NOP
	CBI	PORTB,CLK		;Lower CLK.
	SBI	ADCSRA,ADSC		;Start a new conversion.

	ADD	PIXSUML,PIXEL		;Add to pixel sum for avg.
	BRCC	SUMOK
	INC	PIXSUMH			;Carry into high byte.

SUMOK:	CP	MINPIX,PIXEL		;Pixel <= MINPIX?
	BRCS	MINOK			;  No:  It's fine.

	MOV	MINPIX,PIXEL		;  Yes: It's the new MINPIX.
	MOV	MINLOC,PIXCNT		;       Save location, too.

MINOK:	CP	PIXEL,MAXPIX		;Pixel >= MAXPIX?		
	BRCS	MAXOK			;  No:  It's fine.

	MOV	MAXPIX,PIXEL		;  Yes: It's the new MAXPIX.
	MOV	MAXLOC,PIXCNT		;       Save location, too.

MAXOK:	MOV	ACC,COMMAND		;Outputting each pixel value?
	ANDI	ACC,0x07
	BRNE	DOBIN

	TST	PIXCNT			;  Yes: Is this the first pixel?
	BRNE	RDYOK

	RCALL	PUTRDY			;         Yes: Signal host we're ready.

RDYOK:	MOV	ACC,PIXEL		;       Send the pixel value to host.
	RCALL	PUTBYTE

DOBIN:	RCALL	BINARY			;Do the binary pixel calculations.

NEXTPIX:	INC	PIXCNT			;Next pixel.
	JBRC	PIXCNT,7,READLP		;Back for another if PIXCNT < 128.

	LSr	SHIFT			;Save last seven binary pixels.
	LDI	XL,PIXELS + 31
	ST	X+,SHIFT
	ST	X+,MINPIX		;Save scan stats.
	ST	X+,MINLOC
	ST	X+,MAXPIX
	ST	X+,MAXLOC
	ROL	PIXSUML			;PIXSUM / 128.
	ROL	PIXSUMH
	ST	X+,PIXSUMH
	CLR	MINLOC
	LDI	ACC,254
	MOV	MAXLOC,ACC

CNDRDY:	SBRS	FLAGS,XEQING		;  No:  Executing from buffer?
	RCALL	PUTRDY			;         No:  Signal ready.

JMPMN:	RJMP	MAINLP			;Back for next command.

;--------Binary pixel processor.-------------

BINARY:	TST	PIXCNT			;Starting a new scan?
	BRNE	INSCAN			;  No:  Branch around.

	JBRC	FILTER,FLOAT,FIXED	;  Yes: Floating threshold?

	MOV	_THLD,PIXEL		;         Yes: Initialize _THLD to current pixel.
	CLC				;              Conditionally clear first bit.
	SBRC	THLD,7			;              Is floating offset negative?
	SEC				;                 Yes: Set first bit.
	RJMP	FSTROT

FIXED:	MOV	_THLD,THLD		;         No:  _THLD = THLD always.
	CP	_THLD,PIXEL		;              Set carry iff Pixel > _THLD

FSTROT:	ROr	SHIFT			;       Rotate first comparison into position for hysteresis.
	MOV	pPIXEL,PIXEL		;       Save pPIXEL for first interpolation.
	RJMP	DOTHLD			;       Now go do the real thing.

INSCAN:	ADD	pPIXEL,PIXEL		;Interpolated pixel = (pPIXEL + PIXEL) / 2.
	ROR	pPIXEL
	XCHG	PIXEL,pPIXEL		;Check interpolated pixel first, and set pPIXEL to PIXEL.
	RCALL	DOTHLD			;Call the thresholding routine.
	MOV	PIXEL,pPIXEL		;Now threshold the new pixel.

DOTHLD:	MOV	ACC,_THLD		;Get acting threshold in ACC.
	JBRC	FILTER,FLOAT,OFFSTOK	;Floating threshold?

	ADD	ACC,THLD			;  Yes: Add THLD (i.e. offset)
	JBRC	THLD,7,POSTHLD		;       Is offset negative?

NEGTHLD:	BRCS	OFFSTOK			;         Yes: Carry set?
	CLR	ACC			;                No:  Negative underflow: saturate to zero.
	RJMP	OFFSTOK

POSTHLD:	BRCC	OFFSTOK			;         No:  Carry clear?
	SER	ACC			;                No:  Positive overflow: saturate to 255.

OFFSTOK:	JBRS	FILTER,WINDOW,DOWIN	;Windowing threshold?

	JBRC	SHIFT,7,WASLO		;  No:  Was previous pixel low?

WASHI:	SUB	ACC,HYST			;         No:  High, so subtract hysteresis from thld.
	BRCC	SUBTHLD			;              Saturate to zero if under zero.
	CLR	ACC
	RJMP	SUBTHLD			;              Now go subtract threshold.
	
WASLO:	ADD	ACC,HYST			;         Yes: Low, so add hysteresis to thld.
	BRCC	SUBTHLD			;              Saturate to 255 if over 255.
	SER	ACC

SUBTHLD:	CP	ACC,PIXEL		;       Compare threshold to pixel. "1" bit if PIXEL > thld.
	RJMP	DOSHIFT			;       Go shift it into SHIFT.

DOWIN:	ADD	ACC,HYST			;  Yes: Windowing. Top thld is _THLD + HYST
	BRCS	TRYLOW			;       If over 255, pixel can't be above window.
	CP	ACC,PIXEL		;       Is PIXEL > _THLD + HYST ?
	BRCS	DOSHIFT			;         Yes: So it's outside the window, and carry is set.

TRYLOW:	MOV	ACC,_THLD		;         No:  Bottom thld is _THLD - HYST
	SUB	ACC,HYST	
	BRCS	HIGHNUF
	CLR	ACC			;              Saturate to zero if < zero.
HIGHNUF:	CP	PIXEL,ACC		;       Is PIXEL < _THLD - HYST ?

DOSHIFT:	ROr	SHIFT			;Rotate comparison bit into shift register.
	JBRC	FILTER,FLOAT,PUTPIX	;Using floating threshold?

	MOV	ACC,PIXEL		;_THLD += (PIXEL - _THLD) >> FILTER
	SUB	ACC,_THLD
	MOV	BITCNT,FILTER
	ANDI	BITCNT,0x07

FILTLP:	BREQ	GOTFILT
	INC	ACC
	BREQ	PUTPIX
	DEC	ACC
	BREQ	PUTPIX
	ASR	ACC
	DEC	BITCNT
	RJMP	FILTLP

GOTFILT:	ADD	_THLD,ACC

PUTPIX:	SUBI	XL,0x20			;On an 8-bit boundary?
	BRCC	BINDONE			;  No:  Skip around save to RAM.

	SUBI	XL,-PIXELS-0x20		;  Yes: Get proper address in Z.
	LD	ACC,Z			;       Get byte from previous scan.
	SBRC	COMMAND,NOTPIX		;       Are we NOTting it first?
	COM	ACC			;         Yes: Complement it.
	JBRS	COMMAND,ORPIX,DOOR	;       Is this an OR or XOR?

	SBRS	COMMAND,ANDPIX		;         No:  Is it an AND?
	MOV	ACC,SHIFT		;                No:  Just replace old value with new value.
	SBRC	COMMAND,ANDPIX
	AND	ACC,SHIFT		;                Yes: AND new and old pixels.
	RJMP	STRPIX			;              Go store the result.

DOOR:	SBRS	COMMAND,ANDPIX		;         Yes:  XOR?
	OR	ACC,SHIFT		;                 No:  OR new and old pixelx.
	SBRC	COMMAND,ANDPIX
	EOR	ACC,SHIFT		;                 Yes: XOR new and old pixels.

STRPIX:	ST	X+,ACC			;       Save the result.
	ORI	XL,0xE0			;       Re-init bit counter in upper 3 bits of XL.

BINDONE:	RET				;Done processing binary pixel.


;--------I/O Routines.-----------------------

GETNBUF:	RCALL	GETBYTE			;Get byte from OWIO and buffer.

BUFCMD:	CPI	YL,MAXRAM		;Room to buffer?
	BRCC	TOOCMD			;  No:  Go flag error.

	ST	Y+,ACC			;  Yes: Do it.
	RET

BUFDAT:	CPI	XL,MAXRAM		;Buffer data in ACC. Room?
	BRCC	TOODAT			;  No: Go flag error.

	ST	X+,ACC			;  Yes: Do it.
	RET

TOODAT:	ORI	FLAGS,1<<DATOVF		;Too much data. Set flag.
	RJMP	ERROR			;Go to error routine.

TOOCMD:	ORI	FLAGS,1<<CMDOVF		;Too many commands. Set flag.
	RJMP	ERROR			;Go to error routine.

NOTACMD:	ORI	FLAGS,1<<BADCMD		;Invalid command. Set flag.

ERROR:	RCALL	LOWTIME			;Error routine: Wait for reset.
	RJMP	ERROR

GETNEXT:	JBRC	FLAGS,XEQING,GETBYTE	;Get next command byte. From buffer?
	LD	ACC,Y+			;  Yes: Get it from there.
	RET

GETBYTE:	LDI	BITCNT,8			;Get byte (ACC) from OWIO.
	RJMP	GETBITS

GETBIT:	LDI	BITCNT,1			;Get bit (ACC.7) from OWIO.

	
GETBITS:	RCALL	LOWTIME			;Wait for OWIO high or reset.
	CLR	ACC			;Clear value in case of just one bit.
GB_WT0:	JBIS	PINB,OWIO,GB_WT0		;Wait for OWIO to go low.		
	RCALL	LOWTIME			;Measure the pulse width or do reset.
	CPI	PTIME,64			;Set carry if pulse is narrower than 40uS (i.e. 1)
	ROR	ACC			;Rotate carry into ACC.
GB_ZERO:	DJNZ	BITCNT,GB_WT0		;Loop back for next bit, if any.
	RET				;All done.

;--------------------------------------

PUTBYTE:	LDI	BITCNT,8			;Send byte (ACC) to OWIO.
	RJMP	PUTBITS

PUTRDY:	CLR	ACC			;Send ready (0 bit) to OWIO.
PUTBIT:	LDI	BITCNT,1			;Send bit (ACC.0) to OWIO.

PUTBITS:	RCALL	LOWTIME			;Wait for OWIO high or reset.
PB_WT0:	JBIS	PINB,OWIO,PB_WT0		;Wait for OWIO to go low.
	JBRS	ACC,0,PB_NEXT		;Is next bit a 0?
	SBI	DDRB,OWIO		;  Yes: Assert OWIO output to pull low.
	LDI	PTIME,96			;       Set time for 30 uS.
PB_ZWT:	DJNZ	PTIME,PB_ZWT		;       Wait 96 * 3 clk / 9.6MHz = 30uS.
	CBI	DDRB,OWIO		;       Deassert OWIO output to float high.
PB_NEXT:	RCALL	LOWTIME			;Make sure OWIO actually goes high. May be reset.
	LSR	ACC			;Shift ACC right.
	DJNZ	BITCNT,PB_WT0		;Loop back for next bit, if any.
	RET				;Done.
	
;---------------------------------------	

LOWTIME:	CLR	PTIME			;Measure low width on OWIO in PTIME.
LT_TEST:	SBIC	PINB,OWIO		;Is pin low?		[1 clk]
	RET				;  No:  Return		[1 clk]
	NOP				;			[1 clk]
	IJNZ	PTIME,LT_TEST		;PTIME now > 255?          [3 clk]
LT_WT1:	JBIC	PINB,OWIO,LT_WT1		;  Yes: Low time exceeds 160uS, so reset.
	RJMP	RESET			;       Goto reset after OWIO goes high.

